#!/usr/bin/env rdmd
/*
 *(C) 2018 Dennis Schneider
 * Translate metadata tags from ikiwiki markdown files to Hugo Front-Matter.
 */

module translator;

import std.algorithm;
import std.ascii;
import std.container;
import std.file;
import std.path;
import std.range;
import std.regex;
import std.stdio;
import std.typecons;

auto metaTag = ctRegex!`^\[\[!meta  *([^=]*)="(.*)"\]\]`;
auto tagTag = ctRegex!`^\[\[!tag  *(.*)\]\]`;
auto imageTag = ctRegex!`^\[\[!img  *([^\]]*)\]\]`;
auto linkTag = ctRegex!`\[\[([^|]+)(\|([^\]]+))?\]\]`;

struct Tag {char[] name; char[] value; }

auto separateTags(InRange)(InRange input, Tag[]* pTags) {
  struct filteredRange(InRange) {
    this(InRange _src, Tag[]* _pTags) {
      src = _src; pTags = _pTags;
    }

    bool empty() {
      while(!src.empty()) {
        auto line = src.front;
        auto metaMatch = matchFirst(line, metaTag);
        if (!metaMatch.empty) {
          *pTags ~= Tag(metaMatch[1].dup, metaMatch[2].dup);
          src.popFront();
          continue;
        }

        auto tagMatch = matchFirst(line, tagTag);
        if(!tagMatch.empty()) {
          *pTags ~= Tag(cast(char[])"tags", tagMatch[1].dup);
          src.popFront();
          continue;
        }

        break;
      }

      return src.empty();
    }

    void popFront() {
      src.popFront();
    }

    char[] front() {
      return src.front();
    }

    InRange src;
    Tag[]* pTags;
  }

  return filteredRange!InRange(input, pTags);
}

auto extractTitle(InRange)(InRange input, char[]* pTitle) {
  struct filtered(InRange) {
    this(InRange _src, char[]* _pTitle) {
      src = _src; pTitle = _pTitle;
    }

    bool empty() {
      if(isFirstInternalLine) {
        while(!src.empty) {
          auto line = src.front;
          if(!line.length)
            src.popFront();
          else
            break;
        }

        if(!src.empty) {
          auto line = src.front;
          if(line.startsWith("#")) {
            line = line[1..$];
            while (!line.empty && line[0] == ' ')
              line = line[1..$];
            *pTitle = line.dup;
            src.popFront;
          }
        }

        while(!src.empty) {
          auto line = src.front;
          if(!line.length)
            src.popFront();
          else
            break;
        }

        isFirstInternalLine = false;
      }

      return src.empty();
    }

    void popFront() {
      src.popFront();
    }

    char[] front() {
      return src.front();
    }

    InRange src;
    bool isFirstInternalLine = true;
    char[]* pTitle;
  }

  return filtered!InRange(input, pTitle);
}



struct HugoTags {
  Tag[] tags;
  bool isOk;
}
HugoTags translateTags(Tag[] oldTags) {
  char[] escape(char[] text) {
    if(text.all!(x => isAlphaNum(x) || isWhite(x)))
      return text;
    else
      return "\"" ~ text ~ "\"";
  }

  Tag[] newTags;
  bool isOk = true;

  foreach(tag; oldTags) {
    char[] newValue;
    switch(tag.name) {
    case "date":
      newValue = tag.value.split(' ').take(2).join(' '); 
      break;
    case "keywords":
    case "tags":
      newValue = "\n  - " ~ (
				tag.value
        .split(' ')
        .map!(x => x.strip(','))
        .filter!(x => !x.empty)
        .join("\n  - "));
      break;
    case "title":
      newValue = escape(tag.value);
      break;
    default:
      isOk = false;
      stderr.writefln("Unrecognised tag: %s", tag.name);
    }

    newTags ~= Tag(tag.name, newValue);
  }

  return HugoTags(newTags.nullable, isOk);
}

char[] getValue(char[] text, char[] attribute) {
	auto start = text.find(attribute ~ "=");
	if(start.length < attribute.length + 2)
		return [];

	auto value = start[attribute.length + 1 .. $];
	if(value[0] == '"')
		return value[1..$-value[1..$].find('"').length];
	else if(value[0] == '\'')
		return value[1..$-value[1..$].find('\'').length];
	else
		return value[0..$-value.find(' ').length];
}

char[] convertImageTags(Captures!(char[]) imgTag) {
	auto content = imgTag[1];
	auto nextTag = content.find(' ');
	auto file = content[0..$-nextTag.length];
	auto caption = getValue(content, cast(char[])"caption");
  auto title = getValue(content, cast(char[])"alt");
  auto height = getValue(content, cast(char[])"height");
  auto width = getValue(content, cast(char[])"width");
  auto alignValue = getValue(content, cast(char[])"align");

  switch(alignValue) {
  default:
    alignValue = cast(char[])"right";
    break;
  case "none":
    alignValue = [];
    break;
  case "left":
    alignValue = cast(char[])"left";
    break;
  }

  if(height.empty && width.empty) {
    auto size = getValue(content, cast(char[])"size");
    auto parts = size.split("x");
    width = parts[0];
    height = parts[1];
  }
	
  if (caption.empty)
    caption = title;

	auto hugoTag = "{{< figure src=\"" ~ file ~ "\"";
  if(!caption.empty)
    hugoTag ~= " caption=\"" ~ caption ~ "\"";
  if(!height.empty)
    hugoTag ~= " height=\"" ~ height ~ "\"";
  if(!width.empty)
    hugoTag ~= " width=\"" ~ width ~ "\"";
  if(!alignValue.empty)
    hugoTag ~= " class=\"float-" ~ alignValue ~ "\"";
  hugoTag ~= " >}}";
	return hugoTag;
}

char[] convertInternalLinks(Captures!(char[]) match) {
	return match[3].empty
		? "[" ~ match[1] ~ "]("~ match[1] ~ ")"
		: "[" ~ match[1] ~ "]("~ match[3] ~ ")";
	//return "[" ~ match[1] ~ "]({{< relref \""~ match[2] ~ ".md\" >}})";
}

bool processMarkdownFile(string ikiFilename, string hugoFilename) {
  auto ikiFile = File(ikiFilename, "r");
  auto hugoFile = File(hugoFilename, "wa");

  Tag[] tags;
  char[] title;
  auto content = new DList!(char[])(
    ikiFile.byLine()
    .separateTags(&tags)
    .extractTitle(&title)
    .map!(line => replaceAll!(convertImageTags)(line, imageTag))
    .map!(line => replaceAll!(convertInternalLinks)(line, linkTag))
    .map!dup);
  auto hugoTagsAndFlag = translateTags(tags);
  auto hugoTags = hugoTagsAndFlag.tags;

  if(!title.empty && !hugoTags.any!(x => x.name == "title"))
    hugoTags = Tag(cast(char[])"title",  title) ~ hugoTags;

  hugoFile.writeln("---");
  foreach(tag; hugoTags)
    hugoFile.writeln(tag.name, ": ", tag.value);
  hugoFile.writeln("---");
  if(!content.empty && !content.front.empty)
    hugoFile.writeln("");
  foreach(line; content.opSlice())
    hugoFile.writeln(line);

  return hugoTagsAndFlag.isOk;
}


int main(string[] args) {
  if(args.length != 3) {
    stderr.writefln("Usage: %s old-dir new-dir", args[0]);
    return 1;
  }
  auto oldDir = args[1];
  auto newDir = args[2];

	if(!exists(newDir))
		mkdir(newDir);

  bool ok = true;
  foreach(entry; std.file.dirEntries(oldDir, SpanMode.breadth)) {
    auto oldFile = entry.name;
    if(!oldFile.startsWith(oldDir)) {
      stderr.writeln("Skipping unexpected path: ", entry.name);
      ok = false;
      continue;
    }

    auto newFile = newDir ~ dirSeparator ~ oldFile[oldDir.length..$];
    if(oldFile.isDir) {
      if(!newFile.exists)
        mkdir(newFile);
    } else if (oldFile.isFile) {
      if(oldFile.endsWith(".mdwn")) {
				/*
				if(exists(oldFile[0..$-5]) && isDir(oldFile[0..$-5])) {
					if(!newFile[0..$-5].exists)
						mkdir(newFile[0..$-5]);
					newFile = newFile[0..$-5] ~ "/index.md";
				} else*/
					newFile = newFile[0..$-5] ~ ".md";
        ok &= processMarkdownFile(oldFile, newFile);
      } else
        copy(oldFile, newFile, Yes.preserveAttributes);
    } else {
      stderr.writeln("Warning: Skipping unsupported file ", oldFile);
    }
  }

  return ok ? 0 : 1;
}
